from ..base import safe_import


def instrument(app):
    module = safe_import(
        "opentelemetry.instrumentation.fastapi",
        "pip install acenite-agent[fastapi]",
    )

    FastAPIInstrumentor = module.FastAPIInstrumentor
    FastAPIInstrumentor().instrument_app(app)



from ..base import safe_import


def instrument():
    module = safe_import(
        "opentelemetry.instrumentation.psycopg2",
        "pip install acenite-agent[psycopg2]",
    )

    Psycopg2Instrumentor = module.Psycopg2Instrumentor
    Psycopg2Instrumentor().instrument()



ACENITE_URL   = "http://localhost:8000/server"

ALLOWED_FRAMEWORKS = {
    "flask",
    "fastapi",
    "django",
}

ALLOWED_INSTRUMENTATIONS = {
    "requests",
    "httpx",
    "psycopg2",
    "sqlalchemy",
}



import importlib

from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter

from .constants import ACENITE_URL, ALLOWED_FRAMEWORKS, ALLOWED_INSTRUMENTATIONS
from .integrations import FRAMEWORKS, INSTRUMENTATIONS


def _load_callable(path: str):
    module_path, func_name = path.split(":")
    module = importlib.import_module(module_path)
    return getattr(module, func_name)

def setup_otel(
        *,
        app: object | None,
        framework: str,
        instrumentations: list[str] | None,
        api_key: str,
        service_name: str
) -> None:
    provider = TracerProvider(
        resource=Resource.create({"service.name": service_name})
    )
    trace.set_tracer_provider(provider)

    exporter = OTLPSpanExporter(
                endpoint=f"{ACENITE_URL}/monitor/",
                headers={"Authorization": f"Bearer {api_key}"}
                )

    provider.add_span_processor(BatchSpanProcessor(exporter))

    # -------- Apply framework instrumentation --------
    if framework:
        if framework not in ALLOWED_FRAMEWORKS:
            raise ValueError(f"Unsupported framework: {framework}")

        if framework in ("flask", "fastapi") and app is None:
            raise ValueError(f"{framework} framework requires app instance")

        fn = _load_callable(FRAMEWORKS[framework])

        if framework in ("flask", "fastapi"):
            fn(app)
        else:
            fn()

    # -------- Apply selected instrumentations --------
    if instrumentations:
        for name in instrumentations:
            if name not in ALLOWED_INSTRUMENTATIONS:
                raise ValueError(f"Unsupported instrumentation: {name}")

            fn = _load_callable(INSTRUMENTATIONS[name])
            fn()



from ..base import safe_import


def instrument():
    module = safe_import(
        "opentelemetry.instrumentation.django",
        "pip install acenite-agent[django]",
    )

    DjangoInstrumentor = module.DjangoInstrumentor
    DjangoInstrumentor().instrument()



FRAMEWORKS = {
    "flask": "acenite_agent.integrations.frameworks.flask:instrument",
    "fastapi": "acenite_agent.integrations.frameworks.fastapi:instrument",
    "django": "acenite_agent.integrations.frameworks.django:instrument",
}

INSTRUMENTATIONS = {
    "requests": "acenite_agent.integrations.instrumentations.requests:instrument",
    "httpx": "acenite_agent.integrations.instrumentations.httpx:instrument",
    "psycopg2": "acenite_agent.integrations.instrumentations.psycopg2:instrument",
    "sqlalchemy": "acenite_agent.integrations.instrumentations.sqlalchemy:instrument",
}



from .otel import setup_otel
from .heartbeat import send_heartbeat
from flask import Flask
from apscheduler.schedulers.background import BackgroundScheduler
from opentelemetry import trace


class AceniteAgent:
    _started = False

    @classmethod
    def start(
        cls,
        *,
        app: object | None = None,
        framework: str,
        instrumentations: list[str] | None = None ,
        api_key: str,
        service_name: str = "unknown-service",
        enable_logging: bool = True,
        enable_heartbeat: bool = True,
        heartbeat_interval: float = 60.0,
    ) -> None:

        if cls._started:
            return

        if enable_logging:
            setup_otel(
                app=app,
                framework=framework,
                instrumentations=instrumentations,
                api_key=api_key,
                service_name=service_name
            )

        if enable_heartbeat:
            scheduler = BackgroundScheduler(daemon=True)
            scheduler.add_job(send_heartbeat, "interval", seconds=heartbeat_interval, max_instances=1, coalesce=True,
                kwargs={
                    "api_key": api_key,
                    "interval": heartbeat_interval,
                },
            )
            scheduler.start()

        cls._started = True

    @classmethod
    def get_tracer(cls):
        tracer = trace.get_tracer(__name__)
        return tracer



from ..base import safe_import


def instrument(app):
    module = safe_import(
        "opentelemetry.instrumentation.flask",
        "pip install acenite-agent[flask]",
    )

    FlaskInstrumentor = module.FlaskInstrumentor
    FlaskInstrumentor().instrument_app(app)



from importlib import import_module


class IntegrationError(RuntimeError):
    pass


def safe_import(module_path: str, install_hint: str):
    try:
        return import_module(module_path)
    except Exception as e:
        raise IntegrationError(
            f"Required dependency not installed.\n"
            f"Install with: {install_hint}"
        ) from e



from ..base import safe_import


def instrument():
    module = safe_import(
        "opentelemetry.instrumentation.requests",
        "pip install acenite-agent[requests-trace]",
    )

    RequestsInstrumentor = module.RequestsInstrumentor
    RequestsInstrumentor().instrument()



from ..base import safe_import


def instrument():
    module = safe_import(
        "opentelemetry.instrumentation.httpx",
        "pip install acenite-agent[httpx]",
    )

    HTTPXClientInstrumentor = module.HTTPXClientInstrumentor
    HTTPXClientInstrumentor().instrument()



from ..base import safe_import


def instrument(engine=None):
    module = safe_import(
        "opentelemetry.instrumentation.sqlalchemy",
        "pip install acenite-agent[sqlalchemy]",
    )

    SQLAlchemyInstrumentor = module.SQLAlchemyInstrumentor

    if engine is not None:
        SQLAlchemyInstrumentor().instrument(engine=engine)
    else:
        SQLAlchemyInstrumentor().instrument()



import time, random, uuid, requests

from .constants import ACENITE_URL

BOOT_ID = str(uuid.uuid4())

def send_heartbeat(*,api_key: str, interval: float):
    # jitter each run (avoid thundering herd)
    time.sleep(random.uniform(0.0, interval/10))  # for 60s interval, 0-10% jitter

    try:
        requests.post(
            f"{ACENITE_URL}/heartbeat/",
            headers={"Authorization": f"Bearer {api_key}"},
            json={
                "status": "up",
                "boot_id": BOOT_ID,
                "instance_id": "default",
            },
            timeout=(3, 5),  # connect, read
        )
    except Exception as e:
        # swallow; next run will try again
        print(e)
        pass
